﻿// <summary>ソースコード：ＨＹＭＣＯ演算モデルクラス</summary>
// <author>CommonMP</author>

using System;
using System.Collections.Generic;
using System.Collections;
using System.Text;

using CommonMP.HYSSOP.CoreImpl.HSData;
using CommonMP.HYSSOP.CoreImpl.HSTools;

using CommonMP.HYMCO.Interface.Model;
using CommonMP.HYMCO.Interface.Data;
using CommonMP.HYMCO.CoreImpl.Data;
using CommonMP.HYMCO.CoreImpl.Tool;

// ToDo namespace は　モデル開発者が変更して、ユニークな名称にして下さい
namespace jp.or.jccajsce.StorageFunctionRiver
{
    /// <summary><para>class outline:</para>
    /// <para>演算モデル型紙</para>
    /// </summary>
    /// <remarks><para>history:</para>
    /// <para>[CommonMP][ver 1.0.0][2008/10/01][新規作成]</para>
    /// </remarks>
    public class StorageFunctionRiver : McStateCalModelBase
    {
        /// <summary> 演算データ（キャスト用） </summary>
        StorageFunctionRiverCalInfo m_csSFRInf = null;   // 便利の為、キャスト用に定義しておく

        //=========================
        // 演算実行前処理関連メソッド
        //=========================

        /// <summary><para>method outline:</para>
        /// <para>入力側の接続情報チェック</para>
        /// </summary>
        /// <example>
        /// <para>usage:</para>
        /// <para>bool bRtn = ReceiveConnectionCheck(ref csErrorInf, lInputDataNum, csInputCellData)</para>
        /// </example>
        /// <param name="csErrorInf">エラー出力</param>
        /// <param name="lInputDataNum">入力情報数</param>
        /// <param name="csInputCellData">演算に必要な入力情報配列</param>
        /// <returns>true :正常 , false:異常</returns>
        /// <exception cref="">無し</exception>
        /// <remarks><para>remarks:</para>
        /// <para>受信するデータが自モデルが期待している情報か否かをチェックする</para>
        /// </remarks>
        protected override bool ReceiveConnectionCheck(ref McStructErrorInfo csErrorInf, long lInputDataNum, McReceiveCellDataIF[] csInputCellData)
        {
            bool bRtn = true;

            // 入力端子数のチェック
            if (lInputDataNum == 0)
            {
                csErrorInf.AddCheckErrorData(this.m_csElement.GetID(), StorageFunctionRiverDefine.SFR_MODEL_KIND, "No Receive Port");
                bRtn = false;
            }

            for (long lLp = 0; lLp < lInputDataNum; lLp++)
            {   // 入力する伝送データ数分繰り返します。

                // To Do
                // 入力する伝送データの型式や設定値等をチェックし、一つでも期待情報でなければ、bRtn=false; を設定します。
                // m_csInputData[lLp]
                
                if(StorageFunctionRiverDefine.INFLOW.Equals(csInputCellData[lLp].GetReceivePatternID()))
                {   // 1次元時系列の場合
                    bRtn = true;

                    long lD1 = 0; long lD2 = 0; long lD3 = 0; long lCellDataNum = 0;
                    long lCellDim = csInputCellData[lLp].GetDimension(ref lD1, ref lD2, ref lD3, ref lCellDataNum);
                    if (lD1 < 1)
                    {
                        // エラー
                        bRtn = false;
                        csErrorInf.AddCheckErrorData(this.GetID(), StorageFunctionRiverDefine.SFR_MODEL_KIND,
                             "Cell Dimension is too short. (Send To " + csInputCellData[lLp].GetUpperElementID().ToString() + ")");
                    }
                }
                else if(StorageFunctionRiverDefine.INFLOW_P.Equals(csInputCellData[lLp].GetReceivePatternID()))
                {   // ポイント時系列の場合
                    bRtn = true;
                }
                //long lD1 = 0; long lD2 = 0; long lD3 = 0; long lCellDataNum = 0;
                //long lCellDim = csInputCellData[lLp].GetDimension(ref lD1, ref lD2, ref lD3, ref lCellDataNum);
                //m_csSFRInf.m_lDim = lD1;
                //if (lCellDim == 1 && lD1 >= 1)
                //{   // ポイント時系列または1次元時系列の場合
                //    bRtn = true;
                //}
                else
                {
                    bRtn = false;
                    csErrorInf.AddCheckErrorData(this.m_csElement.GetID(), StorageFunctionRiverDefine.SFR_MODEL_KIND
                        , "Unexpected receive data type. (Received from " + m_csInputData[0].GetUpperElementID().ToString() + ")");
                }
            }
            return bRtn;
        }

        /// <summary><para>method outline:</para>
        /// <para>出力側の接続情報チェック</para>
        /// </summary>
        /// <example>
        /// <para>usage:</para>
        /// <para>bool bRtn = SendConnectionCheck(ref csErrorInf, lOutputDataNum, csOutputCellData)</para>
        /// </example>
        /// <param name="csErrorInf">エラー出力</param>
        /// <param name="lOutputDataNum">出力情報数</param>
        /// <param name="csOutputCellData">出力情報配列</param>
        /// <returns>true :正常 , false:異常</returns>
        /// <exception cref="">無し</exception>
        /// <remarks><para>remarks:</para>
        /// <para>送信端子に設定されている伝送データが自モデルが期待している情報か否かをチェックする</para>
        /// </remarks>
        protected override bool SendConnectionCheck(ref McStructErrorInfo csErrorInf, long lOutputDataNum, McSendCellDataIF[] csOutputCellData)
        {
            bool bRtn = true;

            if (lOutputDataNum == 0)
            {
                // 警告（エラーではない）
                csErrorInf.AddCheckWarningData(this.GetID(), StorageFunctionRiverDefine.SFR_MODEL_KIND, "No send port.");
            }

            for (long lLp = 0; lLp < lOutputDataNum; lLp++)
            {   // 出力する伝送データ数分繰り返します。

                
                // To Do
                // 出力する伝送データの型式や設定値等をチェックし、一つでも期待情報でなければ、bRtn=false; を設定します。
                //m_csOutputData[lLp]
                if (StorageFunctionRiverDefine.OUTFLOW.Equals(csOutputCellData[lLp].GetSendPatternID()))
                {   // 1次元時系列の場合
                    m_csSFRInf.m_lOutPutPortNum += 1;
                    long lD1 = 0; long lD2 = 0; long lD3 = 0; long lCellDataNum = 0;
                    long lCellDim = csOutputCellData[lLp].GetDimension(ref lD1, ref lD2, ref lD3, ref lCellDataNum);
                    if (lD1 < 1)
                    {
                        // エラー
                        bRtn = false;
                        csErrorInf.AddCheckErrorData(this.GetID(), StorageFunctionRiverDefine.SFR_MODEL_KIND,
                             "Cell Dimension is too short. (Send To " + csOutputCellData[lLp].GetLowerElementID().ToString() + ")");
                    }
                }
                else if (StorageFunctionRiverDefine.OUTFLOW_P.Equals(csOutputCellData[lLp].GetSendPatternID()))
                {   // ポイント時系列の場合
                    bRtn = true;
                }
                else
                {
                    // エラー
                    bRtn = false;
                    csErrorInf.AddCheckErrorData(this.GetID(), StorageFunctionRiverDefine.SFR_MODEL_KIND,
                        "Unexpected send data type.");
                }
            }
            return bRtn;
        }

        /// <summary><para>method outline:</para>
        /// <para>モデルを初期化する</para>
        /// </summary>
        /// <example>
        /// <para>usage:</para>
        /// <para>Initialize(csInitialData, lInputDataNum, csInputCellData)</para>
        /// </example>
        /// <param name="csInitialData">初期化設定情報</param>
        /// <param name="lInputDataNum">入力情報数</param>
        /// <param name="csInputCellData">演算に必要な入力情報配列</param>
        /// <returns>true :正常 , false:異常</returns>
        /// <exception cref="">無し</exception>
        /// <remarks><para>remarks:</para>
        /// <para>無し</para>
        /// </remarks>
        protected override bool Initialize(ref McPropertyInfoRoot csInitialData, long lInputDataNum, ref McReceiveCellDataIF[] csInputCellData)
        {
            bool bRtn = false;
            // 引数で与えられたデータを　キャストしています。
            McInitialInfo csInDt = csInitialData as McInitialInfo;

            if (csInDt != null)
            {
                bRtn = true;
                // To Do
                // 初期化が必要な場合、引数によって指定された情報で初期化する
                double dInitFlow_Out = 0d;
                double dInitFlow_In = 0d;
                csInDt.GetInfo("m_dInitFlow_In" , ref dInitFlow_In);
                csInDt.GetInfo("m_dInitFlow_Out" , ref dInitFlow_Out);

                //m_csSFRInf.m_dQoo = dInitFlow;
                m_csSFRInf.m_dQin = dInitFlow_In;
                m_csSFRInf.m_dQoutp = dInitFlow_Out;
                m_csSFRInf.m_dQout = dInitFlow_Out;
            }

            //格納用変数を初期化
            m_csSFRInf.m_hashQout = new SortedList<string , double>();
            //m_csSFRInf.m_dQootlz = 0;
            //m_csSFRInf.m_dQin = 0;

            //シミュレーション開始期間をセット
            m_csSFRInf.m_csLastDataOutTime = new HySTime(0);
            m_csSFRInf.m_csLastDataOutTime.SetTime(this.m_csSimTime);

            return bRtn;
        }

        /// <summary><para>method outline:</para>
        /// <para>モデルを初期化する</para>
        /// </summary>
        /// <example><para>usage:</para>
        /// <para>OnlineInitialize(csInitialData)</para>
        /// </example>
        /// <param name="csInitialData">初期化設定情報</param>
        /// <param name="lInputDataNum">入力情報数</param>
        /// <param name="csInputCellData">演算に必要な入力情報配列</param>
        /// <returns>bool</returns>
        /// <exception cref="">無し</exception>
        /// <remarks><para>remarks:</para>
        /// <para>無し</para>
        /// </remarks>
        protected override bool OnlineInitialize(ref McPropertyInfoRoot csInitialData, long lInputDataNum, ref McReceiveCellDataIF[] csInputCellData)
        {
            return true;
        }

        //=======================
        // 演算実行処理関連メソッド
        //=======================

        /// <summary><para>method outline:</para>
        /// <para>モデル演算</para>
        /// </summary>
        /// <example>
        /// <para>usage:</para>
        /// <para>long lRtn = Calculate(lInputDataNum, ref csInputCellData)</para>
        /// </example>
        /// <param name="lInputDataNum">入力情報数</param>
        /// <param name="csInputCellData">演算に必要な入力情報配列</param>
        /// <returns>0 :正常 , -1:異常</returns>
        /// <exception cref="">無し</exception>
        /// <remarks><para>remarks:</para>
        /// <para>無し</para>
        /// </remarks>
        protected override long Calculate(long lInputDataNum, ref McReceiveCellDataIF[] csInputCellData)
        {
            //McLog.DebugOut(GetSimulationTime(), GetID(), "MyModel", "Calculate", "in");

            // To Do
            // 演算処理を行います。

            NonLnStfn nlnstn = new NonLnStfn();
            //double d_qinx = 0.0;
            double d_qinxx = 0.0;
            double d_qoutxx = 0.0;
            double d_sxx = 0.0;
            //double d_qoutx = 0.0;

            //-Δtの流量
            //d_qinx = m_csSFRInf.m_dQin;

            //流入量取得
            HySCellData[] csCellData = null;
            for (long lLp = 0; lLp < lInputDataNum; lLp++)
            {
                csCellData = csInputCellData[lLp].GetInterpolatedCellD1();
                d_qinxx += csInputCellData[lLp].Data(csCellData[0], 0);
            }

            //d_qoutxx = m_csSFRInf.m_dQoo;
            //d_sxx = m_csSFRInf.m_dSS;

            //d_qoutx = m_csSFRInf.m_dQoutx;
            int iRet = nlnstn.nlsf(  m_csSFRInf.m_dDT
                                   , m_csSFRInf.m_dSc
                                   , m_csSFRInf.m_dPc
                                   , m_csSFRInf.m_dTl
                                   , m_csSFRInf.m_dQl
                                   , m_csSFRInf.m_lParaNum
                                   , m_csSFRInf.m_dSSp
                                   , m_csSFRInf.m_dQin
                                   , d_qinxx
                                   , m_csSFRInf.m_dQoutp
                                   , ref d_sxx
                                   , ref d_qoutxx );
            if (iRet != 0)
            {
                McLog.DebugOut(this.m_csSimTime, m_csElement.GetID(), "StorageFunctionRiverModel",
                               "Calculate::nlsf", Properties.Resources.STRING_MSG_ERROR1);

                System.Windows.Forms.MessageBox.Show(Properties.Resources.STRING_MSG_ERROR1
                                                     + System.Environment.NewLine + this.GetID().ToString()
                                                     , StorageFunctionRiverDefine.SFR_MODEL_NAME.ToString()
                                                     ,System.Windows.Forms.MessageBoxButtons.OK, System.Windows.Forms.MessageBoxIcon.Error);

                return -1;
            }

            //次の時刻の計算に使用する
            
            m_csSFRInf.m_dQin   = d_qinxx;
            m_csSFRInf.m_dQoutp = d_qoutxx;
            m_csSFRInf.m_dQout   = d_qoutxx;
            m_csSFRInf.m_dSSp    = d_sxx;
            m_csSFRInf.m_dSS    = d_sxx;


            //m_csSFRInf.m_dQoutx = m_csSFRInf.m_dQoo;
            

            //流量を格納しておく
            //m_csSFRInf.m_dQin = d_qinxx;

            //最初のCalculateの場合、Tlzによるずらしまで最初のステップの計算値継続
            if ( m_csSFRInf.m_hashQout.Count == 0 )
            {
                m_csSFRInf.m_dQoutTlz = m_csSFRInf.m_dQout;
            }

            // 前時刻
            string stimeNow = HySCalendar.GetString(m_csSFRInf.m_csLastDataOutTime, HySCalendar.FORMAT.lSW_YEAR).ToString();

            // ずらし時間を取得
            HySTime tmTlz = new HySTime(0);
            tmTlz.SetTime(m_csSFRInf.m_csLastDataOutTime);
            tmTlz.AddTime(m_csSFRInf.m_dTlz * 3600);
            string stimeTlz = HySCalendar.GetString(tmTlz, HySCalendar.FORMAT.lSW_YEAR).ToString();

            // ずらし時間のデータを格納
            m_csSFRInf.m_hashQout[stimeTlz] = d_qoutxx; 

            // ずらし時間分のデータを取得
            if (m_csSFRInf.m_hashQout.ContainsKey(stimeNow))
            {
                //m_csSFRInf.m_dQootlz = (double)objTlz;
                m_csSFRInf.m_dQoutTlz = m_csSFRInf.m_hashQout[stimeNow];
            }
            else
            {   //演算間隔とずらし時間間隔ピッチが合わない場合、次のレコードの値をセット
                DateTime dtmSimTime = DateTime.Parse(stimeNow);
                for (int i = m_csSFRInf.m_hashQout.Count - 2; i >= 0; i--)
                {
                    string sTm = m_csSFRInf.m_hashQout.Keys[i];
                    DateTime dtmCurTime = DateTime.Parse(sTm);
                    if (dtmSimTime > dtmCurTime)
                    {
                        m_csSFRInf.m_dQoutTlz = m_csSFRInf.m_hashQout.Values[i + 1];
                        break;
                    }
                }
            }

            // 時間更新
            double dDltTime = this.m_csDltTime.TotalSec();
            m_csSFRInf.m_csLastDataOutTime.AddTime(dDltTime);

            //McLog.DebugOut(GetSimulationTime(), GetID(), "MyModel", "Calculate", "out");
            return 0;
        }

        /// <summary><para>method outline:</para>
        /// <para>モデル演算結果を外部のエレメントに対して公開する</para>
        /// </summary>
        /// <example>
        /// <para>usage:</para>
        /// <para>long lRtn = DataFusion( lOutputDataNum, ref csOutputCellData)</para>
        /// </example>
        /// <param name="lOutputDataNum">出力情報数</param>
        /// <param name="csOutputCellData">出力情報配列</param>
        /// <returns>0 :正常 , -1:異常</returns>
        /// <exception cref="">無し</exception>
        /// <remarks><para>remarks:</para>
        /// <para>無し</para>
        /// </remarks>
        protected override long DataFusion(long lOutputDataNum, ref McSendCellDataIF[] csOutputCellData)
        {
            HySCellData[] csSndCellData = null;

            //McLog.DebugOut(GetSimulationTime(), GetID(), "MyModel", "DataFusion", "in");
            for (long lLp = 0; lLp < lOutputDataNum; lLp++)
            {   // 出力する伝送データ数分繰り返します。

                // To Do
                // 出力する伝送データの型式や設定値等をチェックし、それに合わせて、伝送データに出力値を設定します。
                //m_csOutputData[lLp]
                csSndCellData = csOutputCellData[lLp].PrepareSendCellD1();

                // 流出量、流出量（Tlz考慮なし）、貯留量
                csSndCellData[0].m_dData[0] = m_csSFRInf.m_dQoutTlz;
                csSndCellData[0].m_dData[1] = m_csSFRInf.m_dQout;
                csSndCellData[0].m_dData[2] = m_csSFRInf.m_dSS;

            }

            //McLog.DebugOut(GetSimulationTime(), GetID(), "MyModel", "DataFusion", "out");
            return 0;
        }


        //====================
        // その他必要なメソッド
        //====================

        /// <summary><para>method outline:</para>
        /// <para>プロパティ情報を設定する</para>
        /// </summary>
        /// <example>
        /// <para>usage:</para>
        /// <para>SetProperty(csCellMdlPropertyInfo)</para>
        /// </example>
        /// <param name="csCellMdlPropertyInfo">セル型プロパティ情報</param>
        /// <returns>true :正常 , false:異常</returns>
        /// <exception cref="">無し</exception>
        /// <remarks><para>remarks:</para>
        /// <para>無し</para>
        /// </remarks>
        public override bool SetProperty(McCellModelPropertyIF csCellMdlPropertyInfo)
        {
            bool bRtn = false;
            // 使用しやすいようにキャストしておく
            m_csSFRInf = (StorageFunctionRiverCalInfo)m_csCalInfo;

            // プロパティ設定
            StorageFunctionRiverPropertyInfo csPrptyInfo = csCellMdlPropertyInfo as StorageFunctionRiverPropertyInfo;
            if (csPrptyInfo != null)
            {
                // 演算ステップ時刻設定
                this.m_csDltTime = new HySTime(csPrptyInfo.GetStepTime());

                bRtn = true;

                // To Do
                // モデルの係数、パラメータ等の設定
                csPrptyInfo.GetInfo("m_lParaNum", ref m_csSFRInf.m_lParaNum);
                csPrptyInfo.GetInfo("m_dTl", ref m_csSFRInf.m_dTl);
                csPrptyInfo.GetInfo("m_dTlz", ref m_csSFRInf.m_dTlz);
                SortedList<string, object> data = new SortedList<string, object>();
                csPrptyInfo.GetInfoSQData(ref data);
                m_csSFRInf.m_dSc = (double[])data["m_dSK"];
                m_csSFRInf.m_dPc = (double[])data["m_dP"];
                m_csSFRInf.m_dQl = (double[])data["m_dQl"];

                /***
                ContinuanceCheck(m_csSFRInf.m_lParaNum, m_csSFRInf.m_dSK, m_csSFRInf.m_dP, ref m_csSFRInf.m_dQl);
                ***/

            }
            m_csSFRInf.m_dDT = m_csDltTime.GetTime() / 3600.0;

            return bRtn;
        }

        /// <summary><para>method outline:</para>
        /// <para>演算実行中断中のプロパティ等情報設定</para>
        /// </summary>
        /// <example><para>usage:</para>
        /// <para>bool bRtn = SetOnlineProperty(csCellMdlPropertyInfo)</para>
        /// </example>
        /// <param name="csCellMdlPropertyInfo">プロパティ情報</param>
        /// <returns>無し</returns>
        /// <exception cref="">無し</exception>
        /// <remarks><para>remarks:</para>
        /// <para>演算中断中に　プロパティ情報を変更した場合にコールされる</para>
        /// </remarks>
        public override bool SetOnlineProperty(McCellModelPropertyIF csCellMdlPropertyInfo)
        {
            bool bRtn = false;
            // プロパティ設定
            McCellModelPropertyInfo csPrptyInfo = csCellMdlPropertyInfo as McCellModelPropertyInfo;
            if (csPrptyInfo != null)
            {
                bRtn = true;
            }
            return bRtn;
        }

        /// <summary><para>method outline:</para>
        /// <para>計算開始時に動作する処理</para>
        /// </summary>
        /// <example>
        /// <para>usage:</para>
        /// <para>bool bRtn = ReadyCalculation( )</para>
        /// </example>
        /// <param name="">無し</param>
        /// <returns>true :正常 , false:異常</returns>
        /// <exception cref="">無し</exception>
        /// <remarks><para>remarks:</para>
        /// <para>計算開始の　最初に１回だけコールされるメソッドです。不要ならば、本メソッドをオーバーライドする必要はありません。</para>
        /// </remarks>
        public override bool ReadyCalculation()
        {
            //McLog.DebugOut(GetSimulationTime(), GetID(), "MyModel", "ReadyCalculation", "in");

            // ToDo
            // 計算開始の　最初に１回だけコールされるメソッドです。不要ならば、本メソッドをオーバーライドする必要はありません。
            // （メソッド自身を削除してください）
            for (long lLp = 0; lLp < m_csSFRInf.m_lParaNum; lLp++)
            {
                if (m_csSFRInf.m_dQout <= m_csSFRInf.m_dQl[lLp])
                {
                    m_csSFRInf.m_dSS = m_csSFRInf.m_dSc[lLp] * Math.Pow(m_csSFRInf.m_dQout, m_csSFRInf.m_dPc[lLp])
                                     - m_csSFRInf.m_dTl * m_csSFRInf.m_dQout;
                    break;
                }
            }
            
            m_csSFRInf.m_dSSp = m_csSFRInf.m_dSS;
            m_csSFRInf.m_dQoutp = m_csSFRInf.m_dQout;

            //McLog.DebugOut(GetSimulationTime(), GetID(), "MyModel", "ReadyCalculation", "out");
            return true;
        }

        /***
        public void ContinuanceCheck(long lnum, double[] dK, double[] dP, ref double[] dQlim)
        {
            double dQlim_Cal;

            for (long i = 1; i < lnum; i++)
            {
                dQlim_Cal = Math.Pow(dK[i-1] / dK[i], 1 / (dP[i] - dP[i-1]));
                if (Math.Abs(Math.Round(dQlim_Cal, 3) - dQlim[i-1]) > 0.001)
                {
                    dQlim[i-1] = dQlim_Cal;
                }
            }

            return;
        }
        ***/


        /// <summary><para>method outline:</para>
        /// <para>計算中断時に動作する処理</para>
        /// </summary>
        /// <example>
        /// <para>usage:</para>
        /// <para>bool bRtn = SuspendCalculation( )</para>
        /// </example>
        /// <param name="">無し</param>
        /// <returns>true :正常 , false:異常</returns>
        /// <exception cref="">無し</exception>
        /// <remarks><para>remarks:</para>
        /// <para>オペレーター操作等により計算中断時に　コールされるメソッドです。不要ならば、本メソッドをオーバーライドする必要はありません。</para>
        /// </remarks>
        public override bool SuspendCalculation()
        {
            //McLog.DebugOut(GetSimulationTime(), GetID(), "MyModel", "SuspendCalculation", "in");

            // ToDo
            // オペレーター操作等により計算中断時に　コールされるメソッドです。不要ならば、本メソッドをオーバーライドする必要はありません。
            // （メソッド自身を削除してください）

            //McLog.DebugOut(GetSimulationTime(), GetID(), "MyModel", "SuspendCalculation", "out");
            return true;
        }

        /// <summary><para>method outline:</para>
        /// <para>計算終了時に動作する処理</para>
        /// </summary>
        /// <example>
        /// <para>usage:</para>
        /// <para>bool bRtn = CompleteCalculation( )</para>
        /// </example>
        /// <param name="">無し</param>
        /// <returns>true :正常 , false:異常</returns>
        /// <exception cref="">無し</exception>
        /// <remarks><para>remarks:</para>
        /// <para>計算終了時　最後の最初に１回だけコールされるメソッドです。不要ならば、本メソッドをオーバーライドする必要はありません。</para>
        /// </remarks>
        public override bool CompleteCalculation()
        {
            //McLog.DebugOut(GetSimulationTime(), GetID(), "MyModel", "CompleteCalculation", "in");

            // ToDo
            // 計算終了時　最後の最初に１回だけコールされるメソッドです。不要ならば、本メソッドをオーバーライドする必要はありません。
            // （メソッド自身を削除してください）

            //McLog.DebugOut(GetSimulationTime(), GetID(), "MyModel", "CompleteCalculation", "out");
            return true;
        }

        /// <summary><para>method outline:</para>
        /// <para>演算用刻み時間を自動的に変更する</para>
        /// </summary>
        /// <example>
        /// <para>usage:</para>
        /// <para>ChangeDeltaTimeAutomatically()</para>
        /// </example>
        /// <param name="">無し</param>
        /// <returns>無し</returns>
        /// <exception cref="">無し</exception>
        /// <remarks><para>remarks:</para>
        /// <para>無し</para>
        /// </remarks>
        public override void ChangeDeltaTimeAutomatically()
        {
            //McLog.DebugOut(GetSimulationTime(), GetID(), "MyModel", "ChangeDeltaTimeAutomatically", "in");

            // ToDo
            // 演算中条件によって　自身のδTを(this.m_csDltTime) を変更するメソッドです。
            // 不要ならば、本メソッドをオーバーライドする必要はありません。（メソッド自身を削除してください）

            //McLog.DebugOut(GetSimulationTime(), GetID(), "MyModel", "ChangeDeltaTimeAutomatically", "out");
        }

    }
    //

    /// <summary>
    /// 貯留関数法（河道）クラス
    /// </summary>
    /// <param name="dtk">Δｔ</param>
    /// <param name="sk">Ｋ</param>
    /// <param name="P">Ｐ</param>
    /// <param name="TL">遅滞時間</param>
    /// <param name="qlim">適応流量</param>
    /// <param name="nlim"></param>
    /// <param name="sx">前時間貯流量</param>
    /// <param name="qinx">前時間流入量</param>
    /// <param name="qinxx">現時間流入量</param>
    /// <param name="qoutx">前時間流出量</param>
    /// <param name="sxx">現時間貯流量</param>
    /// <param name="qoutxx">現時間流出量</param>
    /// <returns>0 :正常 , -1:異常</returns>
    public class NonLnStfn
    {
        public int nlsf(double dtk
                        , double[] sk
                        , double[] p
                        , double tl
                        , double[] qlim
                        , long nlim
                        , double sx
                        , double qinx
                        , double qinxx
                        , double qoutx
                        , ref double sxx
                        , ref double qoutxx)
        {
            StringBuilder sb = new StringBuilder(); //デバッグ用
            long lr2 = 100;      // 収束計算の反復回数上限
            double error = 0.01; // 収束計算の許容誤差(m3/s)
            double s2 = 0.0;
            long lsug = 0;
            double q2 = 0.0;
            double q2d = 0.0;

            double q1 = 0.0;
            double q3 = 0.0;

            long lck = 0;

            double[] SmaxQ = new double[sk.GetLength(0)];
            double[] Smax = new double[sk.GetLength(0)];
            long ll3 = 0;
            double[] SminQ = new double[sk.GetLength(0)];

            //k=0,p=0の場合
            if (Math.Abs(sk[0]) < 0.0001 || Math.Abs(p[0]) < 0.0001)
            {
                qoutxx = qinxx;
                sx = 0.0;
                return 0;
            }

            //tlの有無で場合分け
            if (!(tl > 0))
            {//tlが無い場合
                q1 = 0.0;
                q3 = 2.0 * qinxx;
                q3 = Math.Max(q3, 2.0 * qinx);
                q3 = Math.Max(q3, 2.0 * qoutx);

                stable(ref sx, qoutx, sk, p, tl, ref qlim, nlim, ref lsug);

                Label10:
                lck++;
                if (lck > lr2)
                {
                    Console.WriteLine(sb.ToString());
                    return -1;
                }

                q2 = (q1 + q3) * 0.5;

                //前時間流出量のステージ探索
                long ll2 = 0;
                for (long mm = 0; ll2 < nlim; mm++)
                {
                    ll2 = mm;
                    if (qoutx <= qlim[mm]) { break; }
                }

                //現時間流出量(仮計算値)のステージ探索
                ll3 = 0;
                for (long mm = 0; mm < nlim; mm++)
                {
                    ll3 = mm;
                    if (q2 <= qlim[mm]) { break; }
                }

                stable(ref s2, q2, sk, p, tl, ref qlim, nlim, ref lsug);

                //貯留量不連続値の補正
                double deltas = 0.0;
                if (ll2 != ll3)
                {
                    long ll = Math.Min(ll2, ll3);
                    double ss2 = qtos(qlim[ll], sk[ll2], p[ll2], tl);
                    double ss3 = qtos(qlim[ll], sk[ll3], p[ll3], tl);
                    deltas = ss3 - ss2;
                }
                s2 = s2 - deltas;
                
                q2d = 2.0 * (-s2 + sx) / dtk + qinxx + qinx - qoutx;

                sb.Append(string.Format("lck={0,5}, qinx={1:f10}, qinxx={2:f10}, qoutx={3:f10}, q2d={4:f10}, sx={5:f10}, s2={6:f10}, q1={7:f10}, q2={8:f10}, q3={9:f10}"
                                        , lck, qinx, qinxx, qoutx, q2d, sx, s2, q1, q2, q3) + "\n");

                q2d = Math.Max(q2d, 0.0);
                if (Math.Abs(q2d - q2) <= error) { goto Label20; }
                if (q2d <= q2)
                {
                    q3 = q2;
                }
                else
                {
                    q1 = q2;
                }
                goto Label10;
            }
            else
            {//tlが有る場合
                //Sが極大になるときのQを算出
                for (long i = 0; i < sk.GetLength(0); i++)
                {

                    SminQ[i] = 0.0;
                    if (p[i] > 0.99999 && p[i] < 1.00001)
                    {
                        SmaxQ[i] = 999999.9;
                        SminQ[i] = 0.0;
                    }
                    else if (p[i] > 1.00001)
                    {
                        SmaxQ[i] = 999999.9;
                        SminQ[i] = Math.Pow(tl / sk[i], 1 / (p[i] - 1));
                    }
                    else
                    {
                        SmaxQ[i] = Math.Pow(tl / (sk[i] * p[i]), 1 / (p[i] - 1));
                        SminQ[i] = 0.0;
                    }
                    Smax[i] = sk[i] * Math.Pow(SmaxQ[i], p[i]) - tl * SmaxQ[i];

                }

                q1 = 0.0;
                q3 = 2.0 * qinxx;
                q3 = Math.Max(q3, 2.0 * qinx);
                q3 = Math.Max(q3, 2.0 * qoutx);
                stable(ref sx, qoutx, sk, p, tl, ref qlim, nlim, ref lsug);                //前タイムステップ貯留量を算定(↓)
                long ll4 = 0;
                for (long mm = 0; ll4 < nlim; mm++)
                {
                    ll4 = mm;
                    if (qoutx <= qlim[mm]) { break; }
                }
                if (qoutx > SmaxQ[ll4])
                {
                    sx = Smax[ll4];
                }                                                                          //前タイムステップ貯留量を算定(↑)

                  Label10:
                lck++;
                if (lck > lr2)
                {
                    Console.WriteLine(sb.ToString());
                    return -1;
                }

                q2 = (q1 + q3) * 0.5;
                
                //前時間流出量のステージ探索
                long ll2 = 0;
                for (long mm = 0; ll2 < nlim; mm++)
                {
                    ll2 = mm;
                    if (qoutx <= qlim[mm]) { break; }
                }

                //現時間流出量(仮計算値)のステージ探索
                ll3 = 0;
                for (long mm = 0; mm < nlim; mm++)
                {
                    ll3 = mm;
                    if (q2 <= qlim[mm]) { break; }
                }

                
                if (p[ll3] > 0.99999 && p[ll3] < 1.00001)
                {
                    if (sk[ll3] < tl)
                    {
                        q2d = qinxx;
                        goto Label901;
                    }
                }


                stable(ref s2, q2, sk, p, tl, ref qlim, nlim, ref lsug);
                if (q2 > SmaxQ[ll3]) { s2 = Smax[ll3]; }                                       // 極大値の補正

                //貯留量不連続値の補正
                double deltas = 0.0;
                if (ll2 != ll3)
                {
                    long ll = Math.Min(ll2, ll3);
                    double ss2 = qtos(qlim[ll], sk[ll2], p[ll2], tl);
                    if (qlim[ll] > SmaxQ[ll2]) { ss2 = Smax[ll2]; }                            // 極大値の補正
                    double ss3 = qtos(qlim[ll], sk[ll3], p[ll3], tl);
                    if (qlim[ll] > SmaxQ[ll3]) { ss3 = Smax[ll3]; }                            // 極大値の補正
                    deltas = ss3 - ss2;
                }
                s2 = s2 - deltas;
                
                q2d = 2.0 * (-s2 + sx) / dtk + qinxx + qinx - qoutx;

            Label901:
                if (qinxx < SminQ[ll3]) q2d = qinxx;

                sb.Append(string.Format("lck={0,5}, qinx={1:f10}, qinxx={2:f10}, qoutx={3:f10}, q2d={4:f10}, sx={5:f10}, s2={6:f10}, q1={7:f10}, q2={8:f10}, q3={9:f10}"
                                      , lck, qinx, qinxx, qoutx, q2d, sx, s2, q1, q2, q3) + "\n");

                q2d = Math.Max(q2d, 0.0);
                if (Math.Abs(q2d - q2) <= error)
                {
                    goto Label20;
                }


                if (q2d <= q2)
                {
                    q3 = q2;
                }
                else
                {
                    q1 = q2;
                }
                goto Label10;
            }

        Label20:
            qoutxx = q2;
            sxx = s2;
            qoutxx = (q2 + q2d) * 0.5;
            if ((tl > 0))
            {
                if (qoutxx > SmaxQ[ll3])
                {
                    qoutxx = qinxx;
                    sx = Smax[ll3];
                }
            }


            return 0;

        }

        private void stable(ref double s, double q, double[] sk, double[] p, double tl, ref double[] ql, long nl,
            ref long lsug)
        {
            long ll = 0;
            for (long l = 0; l < nl; l++)
            {
                ll = l;
                if (q <= ql[l]) { break; }
            }
            s = qtos(q, sk[ll], p[ll], tl);
            if (p[ll] <= 0.0)
            {
                lsug = 1;
            }
            else
            {
                lsug = 0;
            }
            return;
        }

        private double qtos(double q, double ssk, double pp, double tl)
        {
            double qt = 0.0;
            if (q > 0.0)
            {
                qt = ssk * Math.Pow(q, pp) - tl * q;
            }
            else
            {
                qt = 0.0;
            }
            return qt;
        }
    }

}
